#ifndef _IDSA_INTERNAL_
#define _IDSA_INTERNAL_

#include <stdarg.h>
#include <idsa.h>

#ifdef WANTS_PROF
#include <sys/times.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif
/* constants *************************************************************** *//*#define IDSA_SOCKET      "/tmp/idsa" */
#define IDSA_SOCKET      "/var/run/idsa"
/* maxima used inside below structures ************************************ */
#define IDSA_S_OFFSET (5*sizeof(unsigned int))
#define IDSA_M_UNITS  (IDSA_M_MESSAGE-IDSA_S_OFFSET)
#define IDSA_M_RISKS         3	/* number of risk types */
/* define the structures mentioned in idsa.h ****************************** */ struct idsa_unit {
    char u_name[IDSA_M_NAME];
    unsigned int u_type;
    char u_ptr[IDSA_M_LONG];
  };

  struct idsa_event {
    unsigned int e_magic;	/* magic */
    unsigned int e_size;	/* size to be written */
    unsigned int e_count;	/* count of units */
    unsigned int e_repeat;	/* number of repeats */
    unsigned int e_result;	/* result: allow, deny */
    char e_ptr[IDSA_M_UNITS];
  };

  struct idsa_connection {
    int c_fd;
    int c_error;
    int c_backoff;
    int c_result;

    unsigned int c_count;
    unsigned int c_flags;	/* flags: failopen */

    IDSA_UNIT *c_fail[IDSA_M_RISKS];	/* on failure allow risks in this range */
    IDSA_UNIT *c_pre[IDSA_M_RISKS];	/* filter out risks in this range */
    double c_scale[IDSA_M_RISKS];	/* scale risks by this factor */

    IDSA_EVENT *c_reply;	/* read reply */
    unsigned int c_length;	/* amount of reply read */

    char c_reason[IDSA_M_MEDIUM];

    char c_service[IDSA_M_MEDIUM];	/* service name - kept for resetting template */
    char c_credential[IDSA_M_MEDIUM];	/* unused */

    IDSA_EVENT *c_template;	/* template from which other events are copied */
    IDSA_EVENT *c_cache;	/* simple cache for malloc */

#ifdef WANTS_PROF
    clock_t c_libtime;
    clock_t c_systime;
    clock_t c_waltime;
    int c_profnum;
#endif
  };

/* manipulation of units  ************************************************* */

  char *idsa_unit_name(IDSA_UNIT * u);
  unsigned int idsa_unit_type(IDSA_UNIT * u);
  int idsa_unit_size(IDSA_UNIT * u);

  IDSA_UNIT *idsa_unit_new(char *n, unsigned int t, char *s);
  IDSA_UNIT *idsa_unit_dup(IDSA_UNIT * u);
  void idsa_unit_copy(IDSA_UNIT * a, IDSA_UNIT * b);
  void idsa_unit_free(IDSA_UNIT * u);

#define IDSA_COMPARE_LESS       0x01	/* a < b */
#define IDSA_COMPARE_EQUAL      0x02	/* a == b */
#define IDSA_COMPARE_INTERSECT  0x03	/* a n b != 0 */
#define IDSA_COMPARE_MORE       0x04	/* a > b */

  typedef int (*IDSA_FUNCTION_COMPARE) (IDSA_UNIT * a, IDSA_UNIT * b);
  typedef int (*IDSA_FUNCTION_CHECK) (IDSA_UNIT * u);
  typedef int (*IDSA_FUNCTION_GET) (IDSA_UNIT * u, void *p, int l);
  typedef int (*IDSA_FUNCTION_SET) (IDSA_UNIT * u, void *p);
  typedef int (*IDSA_FUNCTION_SCAN) (IDSA_UNIT * u, char *s);
  typedef int (*IDSA_FUNCTION_PRINT) (IDSA_UNIT * u, char *s, int l,
				      int m);

  int idsa_unit_compare(IDSA_UNIT * a, IDSA_UNIT * b);
  int idsa_unit_check(IDSA_UNIT * u);
  int idsa_unit_get(IDSA_UNIT * u, void *p, int l);
  int idsa_unit_set(IDSA_UNIT * u, void *p);
  int idsa_unit_scan(IDSA_UNIT * u, char *s);
  int idsa_unit_print(IDSA_UNIT * u, char *s, int l, int m);

/* type related information *********************************************** */

  struct idsa_type_details;
  typedef struct idsa_type_details IDSA_TYPE_DETAILS;
  IDSA_TYPE_DETAILS *idsa_type_lookup(unsigned int t);

  unsigned int idsa_type_code(char *n);
  char *idsa_type_name(unsigned int t);
  unsigned int idsa_type_c(unsigned int t);
  int idsa_type_size(unsigned int t);

/* manipulation of events ************************************************* */

  IDSA_EVENT *idsa_event_new(unsigned int m);
  void idsa_event_clear(IDSA_EVENT * e, unsigned int m);
  void idsa_event_free(IDSA_EVENT * e);
  void idsa_event_copy(IDSA_EVENT * a, IDSA_EVENT * b);
  int idsa_event_check(IDSA_EVENT * e);

  void idsa_event_result(IDSA_EVENT * e, unsigned int r);

  unsigned int idsa_event_unitcount(IDSA_EVENT * e);
  IDSA_UNIT *idsa_event_unitbyname(IDSA_EVENT * e, char *n);
  IDSA_UNIT *idsa_event_unitbynumber(IDSA_EVENT * e, int n);
  IDSA_UNIT *idsa_event_unitappend(IDSA_EVENT * e, char *n,
				   unsigned int t);

  int idsa_event_vset(IDSA_EVENT * e, char *n, char *s, int f, double ar,
		      double cr, double ir, va_list ap);
  int idsa_event_vscan(IDSA_EVENT * e, char *n, char *s, int f, double ar,
		       double cr, double ir, va_list ap);

  int idsa_event_set(IDSA_EVENT * e, char *n, char *s, int f, double ar,
		     double cr, double ir, ...);
  int idsa_event_scan(IDSA_EVENT * e, char *n, char *s, int f, double ar,
		      double cr, double ir, ...);

  IDSA_UNIT *idsa_event_setbynumber(IDSA_EVENT * e, int n, void *p);
  IDSA_UNIT *idsa_event_scanbynumber(IDSA_EVENT * e, int n, char *s);

  IDSA_UNIT *idsa_event_setappend(IDSA_EVENT * e, char *n, unsigned int t,
				  void *p);
  IDSA_UNIT *idsa_event_scanappend(IDSA_EVENT * e, char *n, unsigned int t,
				   char *s);

  int idsa_event_syslog(IDSA_EVENT * e, int pri, char *msg);

#include <stdio.h>

  int idsa_event_dump(IDSA_EVENT * e, FILE * f);

/* risks ******************************************************************* */

#define IDSA_M_RISKTOP          0xfff0	/* highest internal representation */
#define IDSA_RISKDEFAULT        0.9
#define IDSA_RISKSCALE          1.0	/* default risk scale factor */

/* required and reserved units ********************************************* */

#define IDSA_N_PID               0	/* "pid" */
#define IDSA_N_UID               1	/* "uid" */
#define IDSA_N_GID               2	/* "gid" */
#define IDSA_N_TIME              3	/* "time" */
#define IDSA_N_SERVICE           4	/* "service" */
#define IDSA_N_HOST              5	/* "host" */
#define IDSA_N_NAME              6	/* "name" */
#define IDSA_N_SCHEME            7	/* "scheme" */
#define IDSA_N_HONOUR            8	/* "honour" */
#define IDSA_N_ARISK             9	/* "arisk: availability" */
#define IDSA_N_CRISK            10	/* "crisk: confidentiality" */
#define IDSA_N_IRISK            11	/* "irisk: integrity" */

#define IDSA_M_REQUIRED         12	/* fields required + 1 (note M) */

#define IDSA_N_IGNORED          12	/* "ignored" */
#define IDSA_N_JOB              13	/* "job" */
#define IDSA_N_REASON           14	/* "reason" */
#define IDSA_N_COMMENT          15	/* "comment" */
#define IDSA_N_PREARISK         16	/* "pre/arisk" */
#define IDSA_N_PRECRISK         17	/* "pre/crisk" */
#define IDSA_N_PREIRISK         18	/* "pre/irisk" */
#define IDSA_N_FAILARISK        19	/* "fail/arisk" */
#define IDSA_N_FAILCRISK        20	/* "fail/crisk" */
#define IDSA_N_FAILIRISK        21	/* "fail/irisk" */

#define IDSA_M_RESERVED         22	/* fields reserved + 1 (note M) */

/* get number of required fields */
  int idsa_required_count();

/* lookup reserved fields */
  unsigned int idsa_reserved_typebynumber(int n);
  unsigned int idsa_reserved_typebyname(char *n);
  char *idsa_reserved_namebynumber(int n);
  int idsa_reserved_numberbyname(char *n);

/* client / server io ***************************************************** */

  int idsa_request_default(IDSA_EVENT * e, char *service, char *group,
			   char *name);
  int idsa_request_check(IDSA_EVENT * e);

  int idsa_reply_default(IDSA_EVENT * e);
  int idsa_reply_check(IDSA_EVENT * e);

  int idsa_event_tobuffer(IDSA_EVENT * e, char *s, int l);
  int idsa_event_frombuffer(IDSA_EVENT * e, char *s, int l);

/* assorted output formats ************************************************ */

  typedef int (*IDSA_PRINT_FUNCTION) (IDSA_EVENT * e, char *b, int l);

  struct idsa_print_format;
  typedef struct idsa_print_format IDSA_PRINT_FORMAT;

  IDSA_PRINT_FORMAT *idsa_print_lookup(char *n);
  int idsa_print(IDSA_EVENT * e, char *b, int l, IDSA_PRINT_FORMAT * p);

  int idsa_print_native(IDSA_EVENT * e, char *b, int l);
  int idsa_print_internal(IDSA_EVENT * e, char *b, int l);
  int idsa_print_syslog(IDSA_EVENT * e, char *b, int l);
  int idsa_print_ulm(IDSA_EVENT * e, char *b, int l);
  int idsa_print_tulm(IDSA_EVENT * e, char *b, int l);
  int idsa_print_csv(IDSA_EVENT * e, char *b, int l);

/* profiling stuff ******************************************************** */

#ifdef WANTS_PROF
  int idsa_prof(IDSA_CONNECTION * c, FILE * fp);
#endif

#ifdef __cplusplus
}
#endif
#endif
