#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <unistd.h>
#include <sched.h>

#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>

#include <idsa.h>

#include "udomain.h"
#include "misc.h"

#include "idsad.h"
#include "job.h"
#include "rule.h"
#include "parse.h"
#include "set.h"
#include "messages.h"

static void handle(int s);
static void parent(int fd, pid_t pid);

volatile int signum = 0;

void usage()
{
  printf("idsad %s\n", VERSION);
  printf
      ("Usage: idsad [-knuv] [-f file] [-i username] [-p socket ...] [-r directory]\n");
  printf
      ("-f file          use alternate configuration file (default is %s)\n",
       IDSAD_CONFIG);
  printf("-i username      run as this username (no default)\n");
  printf
      ("-k               kill existing idsad instance (instead of lockfile)\n");
  printf("-n               do not fork into background\n");
  printf
      ("-p socket ...    whitespace delimited list of unix domain sockets to listen on\n");
  printf("                 (default is %s)\n", IDSA_SOCKET);
  printf("-r directory     chroot to directory (no default)\n");
  printf("-u               honour umask when creating sockets\n");
  printf("-v               print version\n");

  exit(0);
}

#define COPY_BUFFER 128

static void parent(int fd, pid_t pid)
{
  char buf[COPY_BUFFER];
  int result = 0;
  int rr, status;

  do {
    rr = read(fd, buf, COPY_BUFFER);
    switch (rr) {
    case -1:
      switch (errno) {
      case EAGAIN:
      case EINTR:
	rr = 1;
	break;
      default:
	fprintf(stderr, "idsad: unable to read child messages: %s\n",
		strerror(errno));
	fflush(stderr);
	break;
      }
      break;
    case 0:
      /* eof */
      break;
    default:
      write(STDERR_FILENO, buf, rr);
      /* don't care if write fails, can't do anything about it */
      break;
    }
  } while (rr > 0);

  /* try to collect an exit code in case of failure */
  sched_yield();

#ifdef TRACE
  fprintf(stderr, "parent(): finished\n");
#endif


  if (waitpid(pid, &status, WNOHANG) > 0) {	/* got a child */
    if (WIFEXITED(status)) {
      result = WEXITSTATUS(status);
      fprintf(stderr, "idsad: exited with code %d\n", result);
    } else if (WIFSIGNALED(status)) {
      /* mimic child */
      result = WTERMSIG(status);
      fprintf(stderr, "idsad: killed by signal %d\n", result);
      fflush(stderr);
      raise(result);
    } else if (WIFSTOPPED(status)) {	/* clever dick mode */
      result = WSTOPSIG(status);
      fprintf(stderr, "idsad: stopped by signal %d, restarting with %d\n",
	      result, SIGCONT);
      kill(pid, SIGCONT);
      result = 0;
    }
  }				/* else child probably ok */
  exit(result);
}

static void handle(int s)
{
  if (signum != SIGTERM) {	/* TERM is too important to overwrite */
    signum = s;
  }
}

int main(int argc, char **argv)
{
  int mfd, smfd, sr;		/* variables for select */
  fd_set fsr, fsw;

  int lc, *ltable;		/* listen variables */
  JOB *j, *jtmp;		/* job variables */

  STATE_SET *set;		/* almost all state kept here */

  int run;			/* control mainloop */
  int p[2];			/* pipe to read error messages from child */
  pid_t pid;			/* pid of child */
  pid_t peer;			/* pid of peer to be killed */
  mode_t mask;			/* the mask */
  int status;			/* return status of exited child */
  struct sigaction sag;		/* need fine grained control over signals */

  char *config;			/* name of config file */
  char *rootdir;		/* directory to do a chroot to */
  int nofork;			/* don't fork into background (default: fork) */
  int noumask;			/* disable umask when creating sockets (default: ignore umask) */
  char *id;			/* string containing user name */
  int zap;			/* kill any running instance before starting */

  int i, k, t;			/* misc */

  /* cmdline defaults */
  id = NULL;
  config = NULL;
  rootdir = NULL;
  nofork = 0;
  noumask = 1;
  zap = 0;

  ltable = NULL;
  lc = 0;
  signum = 0;

  /* set to keep -Wall -O2 happy */
  pid = 0;
  mask = 0;
  peer = 0;

  /* nontrivial initialization: */
  /* 1) provisional argument parsing */
  /* 2) possibly fork into background, with parent() watching stderr */
  /* 3) real argument parsing */
  /* 4) fill in defaults */
  /* 5) parse config file */
  /* 6) drop root and chroot */
  /* 7) signal handlers */

  /* provisional parse */
  i = 1;
  k = 1;
  while (i < argc) {
    if (argv[i][0] == '-') {
      switch (argv[i][k]) {
      case 'c':
	printf
	    ("(c) 2000 Marc Welz: Licensed under the terms of the GNU General Public License\n");
	exit(0);
	break;
      case 'h':		/* print brief help message */
	usage();
	break;
      case 'k':		/* kill any existing instance */
	zap++;
	k++;
	break;
      case 'n':		/* keep in foreground */
	nofork++;
	k++;
	break;
      case 'u':
	noumask = 0;
	k++;
	break;
      case 'v':
	printf("idsad %s\n", VERSION);
	exit(0);
	break;
      case '\0':
	k = 1;
	i++;
	break;
      default:
	k++;
	break;
      }
    } else {
      i++;
    }
  }

  /* go into background, get error messages via pipe */
  if (nofork == 0) {
    if (pipe(p)) {
      fprintf(stderr, "idsad: unable to create pipe: %s\n",
	      strerror(errno));
      exit(1);
    }
    fflush(stderr);
    pid = fork();
    switch (pid) {
    case -1:
      fprintf(stderr, "idsad: unable to fork: %s\n", strerror(errno));
      exit(1);
      break;
    case 0:
      close(p[0]);
      if (dup2(p[1], STDERR_FILENO) != STDERR_FILENO) {
	/* ouch */
	write(p[1], "idsad: unable to dup2 file descriptor\n", 38);
	exit(1);
      }
      close(p[1]);
      close(STDOUT_FILENO);
      close(STDIN_FILENO);
      setsid();
      break;
    default:
      close(p[1]);
      parent(p[0], pid);
      break;
    }
  }
  /* don't panic in case other end goes away */
  sag.sa_handler = SIG_IGN;
  sigfillset(&(sag.sa_mask));
  sag.sa_flags = SA_RESTART;
  sigaction(SIGPIPE, &sag, NULL);

  if (noumask) {
    /* clear out umask */
    mask = umask(S_IXUSR | S_IXGRP | S_IXOTH);
  }
  mfd = 0;
  i = 1;
  k = 1;
  while (i < argc) {
    if (argv[i][0] == '-') {
      switch (argv[i][k]) {
      case 'p':
	k++;
	if (argv[i][k] == '\0') {
	  k = 0;
	  i++;
	}
	while ((i < argc) && (argv[i][k] != '-')) {
	  ltable = realloc(ltable, sizeof(int) * (lc + 1));
	  if (ltable) {
	    ltable[lc] = udomainlisten(argv[i] + k, IDSAD_BACKLOG, zap);
	    if (ltable[lc] == (-1)) {
	      fprintf(stderr, "idsad: unable to listen on socket %s: %s\n",
		      argv[i] + k, strerror(errno));
	      exit(1);
	    }
	    if (ltable[lc] > mfd) {
	      mfd = ltable[lc];
	    }
	    lc++;
	    k = 0;
	    i++;
	  } else {
	    fprintf(stderr,
		    "idsad: unable to allocate listen socket table of %d elements\n",
		    lc + 1);
	    exit(1);
	  }

	}
	break;
      case 'i':
	k++;
	if (argv[i][k] == '\0') {
	  k = 0;
	  i++;
	}
	if (i < argc) {
	  id = argv[i] + k;
	  i++;
	  k = 1;
	} else {
	  fprintf(stderr,
		  "idsad: -i option requires a user id as parameter\n");
	  exit(1);
	}
	break;
      case 'f':
	k++;
	if (argv[i][k] == '\0') {
	  k = 0;
	  i++;
	}
	if (i < argc) {
	  config = argv[i] + k;
	  i++;
	  k = 1;
	} else {
	  fprintf(stderr,
		  "idsad: -f option requires a configuration file as parameter\n");
	  exit(1);
	}
	break;
      case 'r':
	k++;
	if (argv[i][k] == '\0') {
	  k = 0;
	  i++;
	}
	if (i < argc) {
	  rootdir = argv[i] + k;
	  i++;
	  k = 1;
	} else {
	  fprintf(stderr,
		  "idsad: -r option requires a directory as parameter\n");
	  exit(1);
	}
	break;
	/* these options have already been handled previously */
      case 'v':
      case 'u':
      case 'n':
      case 'k':
      case 'h':
      case 'c':
      case '-':
	k++;
	break;
      case '\0':
	k = 1;
	i++;
	break;
      default:
	fprintf(stderr, "idsad: unknown option -%c\n", argv[i][k]);
	exit(1);
	break;
      }
    } else {
      fprintf(stderr, "idsad: unknown argument %s\n", argv[i]);
      exit(1);
    }
  }

  /* build set of rules soon since it is most likely cause of errors */
  set = parse_set(config ? config : IDSAD_CONFIG);
  if (set == NULL) {
    fprintf(stderr, "idsad: unable to allocate memory for state\n");
    exit(1);
  } else if (!parse_ok(set)) {	/* config file is broken */
    fprintf(stderr, "idsad: %s\n", set->s_errbuf);
    set_free(set);
    exit(1);
  }
  /* if user has not specified a socket we listen on default */
  if (lc == 0) {
    ltable = malloc(sizeof(int));
    if (ltable) {
      ltable[0] = udomainlisten(IDSA_SOCKET, IDSAD_BACKLOG, zap);
      if (ltable[0] != (-1)) {
	mfd = ltable[0];
	lc = 1;
      } else {
	fprintf(stderr, "idsad: unable to listen on socket %s: %s\n",
		IDSA_SOCKET, strerror(errno));
	exit(1);
      }
    } else {
      fprintf(stderr, "idsad: unable to allocate memory for an integer\n");
      exit(1);
    }
  }
  if (noumask) {
    umask(mask);
  }
  drop_root("idsad", id, rootdir);
  drop_fork("idsad");

  sag.sa_handler = handle;
  sigfillset(&(sag.sa_mask));

  sag.sa_flags = SA_RESTART;	/* minor ones */
  signal(SIGCHLD, handle);
  signal(SIGHUP, handle);

  sag.sa_flags = 0;		/* serious signals */
  signal(SIGINT, handle);
  signal(SIGTERM, handle);
  signal(SIGALRM, handle);

  /* cache our own uid */
  set->s_gid = getgid();

#ifdef TRACE
  fprintf(stderr, "main(): starting up\n");
#endif

#ifndef TRACE
  /* close stderr, parent will notice and go away */
  if (nofork == 0) {
    fflush(stderr);
    fclose(stderr);
  }
#endif

  set_clear(set);

  if (zap) {			/* give previous instance a chance to quit before writing */
    sched_yield();
  }
  message_start(set, VERSION);

  /* enter main loop */
  run = 1;
  do {

    /* all signals indicate an abnormal condition including SIGCHLD */
    /* none of our children should die. Hence we do not worry about */
    /* loosing signals if several delivered */

    switch (signum) {
    case SIGINT:
    case SIGTERM:
#ifdef TRACE
      fprintf(stderr, "main(): received TERM signal\n");
#endif
      run = 0;
      signum = 0;
      /* give us a second before interrupting select */
      alarm(1);
      break;
    case SIGHUP:
      signum = 0;
      message_error_unhandled(set, "sighup");
      break;
    case SIGALRM:		/* ha, this should never happen */
      signum = 0;
      run = 0;
      break;
    case SIGCHLD:
      /* FIXME: maybe report this too ? */
      while (waitpid(-1, &status, WNOHANG) > 0);
      signum = 0;
      break;
    }

    FD_ZERO(&fsr);
    FD_ZERO(&fsw);

    /* listen sockets */
    for (i = 0; i < lc; i++) {
      FD_SET(ltable[i], &fsr);
    }

    /* connected sockets */
    for (i = 0; i < set->s_jobcount; i++) {
      j = &(set->s_jobs[i]);

      FD_SET(j->j_fd, &fsr);	/* add all to read list */

      if (job_iswrite(j)) {	/* add those who have something to write list */
	FD_SET(j->j_fd, &fsw);
      }
    }

    sr = select(mfd + 1, &fsr, &fsw, NULL, NULL);

    if (sr > 0) {

      /* try to accept a new connection */
      for (i = 0; i < lc; i++) {
	if (FD_ISSET(ltable[i], &fsr)) {
	  if (set->s_jobcount >= set->s_jobmax) {	/* make space for new slot if required */
	    jtmp = realloc(set->s_jobs, sizeof(JOB) * 2 * set->s_jobmax);
	    if (jtmp) {
	      set->s_jobs = jtmp;
	      set->s_jobmax *= 2;
	    } else {		/* no space */
	      /* MESSAGE: memory allocation failure */
	      message_error_system(set, errno, "accept_client");
	    }
	  }
	  if (set->s_jobcount < set->s_jobmax) {
	    j = &(set->s_jobs[set->s_jobcount]);
	    if (job_accept(j, ltable[i]) == 0) {
	      /* MESSAGE: job connect */
	      if (message_connect(set, j->j_pid, j->j_uid, j->j_gid)) {	/* instruction to drop connection */
		j->j_state = JOB_STATEFIN;
	      }
	      if (mfd < j->j_fd) {	/* win: update mfd as little as possible ;-) */
		mfd = j->j_fd;
	      }
	      set->s_jobcount++;
	    } else {
	      /* MESSAGE: accept failure */
	      message_error_system(set, errno, "accept_client");
	    }
	  } else {		/* problem, we don't accept, so select comes back immediately - to avoid busy loop we drop */
	    job_drop(ltable[i]);
	  }
	}
      }				/* end of handling listeners */

      /* handle connected sockets */
      for (i = 0; i < set->s_jobcount; i++) {
	j = &(set->s_jobs[i]);

	if (FD_ISSET(j->j_fd, &fsw)) {	/* drain out write buffer */
#ifdef TRACE
	  fprintf(stderr, "main(): write activity on client, fd=<%d>\n",
		  j->j_fd);
#endif
	  job_write(j);
	}
	if (FD_ISSET(j->j_fd, &fsr)) {	/* fill in read buffer */
#ifdef TRACE
	  fprintf(stderr, "main(): read activity on client, fd=<%d>\n",
		  j->j_fd);
#endif
	  job_read(j);
	}
	if (job_iswork(j)) {	/* are we waiting for input and has it arrived ? */
	  job_do(j, set);
	}
	/* shut down connections */
	if (job_isend(j)) {
	  /* message: job disconnect */
	  message_disconnect(set, j->j_pid, j->j_uid, j->j_gid);
	  smfd = j->j_fd;
	  job_end(j);
	  if (i < (set->s_jobcount - 1)) {
	    job_copy(j, &(set->s_jobs[set->s_jobcount - 1]));
	    i--;		/* WARNING: decrement so that copied job gets slice */
	  }
	  if (set->s_jobcount > 0) {
	    set->s_jobcount--;
	  }
	  if (smfd >= mfd) {	/* recompute mfd if we have a reason */

	    mfd = ltable[0];
	    for (t = 1; t < lc; t++) {
	      if (ltable[t] > mfd) {
		mfd = ltable[t];
	      }
	    }
	    for (t = 0; t < set->s_jobcount; t++) {
	      if (set->s_jobs[t].j_fd > mfd) {
		mfd = set->s_jobs[t].j_fd;
	      }
	    }
#ifdef TRACE
	    fprintf(stderr, "main(): reduced mfd from <%d> to <%d>\n",
		    smfd, mfd);
#endif
	  }
	}			/* end of shutdown */
      }				/* handle connected sockets */
    }
    /* select */
    /* report if any problems during logging */
    output_error(set);

  } while (run);

#ifdef TRACE
  fprintf(stderr, "main(): finished\n");
#endif

  /* delete listeners */
  for (i = 0; i < lc; i++) {
    close(ltable[i]);
  }
  free(ltable);
  ltable = NULL;
  lc = 0;

  /* MESSAGE: stop */
  message_stop(set);

  /* delete state set */
  set_free(set);

  return 0;
}
