
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "mex.h"

#define MEX_ERROR_OK      0x00
#define MEX_ERROR_TOOLONG 0x01
#define MEX_ERROR_READ    0x02
#define MEX_ERROR_MEMORY  0x03
#define MEX_ERROR_LEX     0x04

#define MEX_STATE_QUIT    0x00
#define MEX_STATE_EATWS   0x01
#define MEX_STATE_EATHASH 0x02
#define MEX_STATE_WORD    0x03
#define MEX_STATE_STRING  0x04

static MEX_TOKEN *mex_new(int id, int type, char *buf, int len, int line);
static MEX_TOKEN *mex_token(MEX * m);
static void mex_delkeyword(MEX_KEY * k);
static int mex_findword(MEX * m, unsigned char *s);
static void mex_dumpkey(MEX_KEY * k, FILE * f, int p);
static void mex_free(MEX_TOKEN * t);

/* should have a sentinel EOF instead of NULL */

MEX *mex_open(char *fname)
{
  MEX *m = NULL;
  int fd;

#ifdef TRACE
  fprintf(stderr, "mex_open(): opening <%s>\n", fname);
#endif

  fd = open(fname, O_RDONLY);
  if (fd != (-1)) {
    fcntl(fd, F_SETFD, FD_CLOEXEC);
    m = mex_fdopen(fd);
  }
  return m;
}

MEX *mex_fdopen(int fd)
{
  MEX *m;
  int i;

  m = malloc(sizeof(MEX));
  if (m) {
    m->m_token = NULL;
    m->m_garbage = NULL;
    m->m_read = 0;
    m->m_lexed = 0;
    m->m_line = 1;
    m->m_eof = 0;

    m->m_error = MEX_ERROR_OK;
    m->m_fd = fd;

    m->m_keywords = NULL;

    for (i = 0; i < MEX_CHAR; i++) {
      m->m_keychars[i] = 0;
    }
  }
  return m;
}

int mex_close(MEX * m)
{
  int result;
  MEX_TOKEN *t, *u;

  result = m->m_error;

  /* zap tokens not yet gotten */
  t = m->m_token;
  while (t) {
    u = t;
    t = t->t_next;
    mex_free(u);
  }
  m->m_token = NULL;

  /* collect tokens gotten */
  t = m->m_garbage;
  while (t) {
    u = t;
    t = t->t_next;
    mex_free(u);
  }
  m->m_garbage = NULL;

  mex_delkeyword(m->m_keywords);
  m->m_keywords = NULL;

  close(m->m_fd);
  m->m_fd = (-1);
  free(m);

  return result;
}

int mex_keychar(MEX * m, unsigned char c, int k)
{
  m->m_keychars[c] = k;

  return 0;
}

int mex_keyword(MEX * m, unsigned char *s, int k)
{
  int result = 0;
  MEX_KEY *p, *q;
  int c, run;

  p = malloc(sizeof(MEX_KEY));
  if (p) {
    p->k_id = k;
    p->k_left = NULL;
    p->k_right = NULL;
    p->k_name = strdup(s);
    if (p->k_name) {

      if (m->m_keywords == NULL) {
	m->m_keywords = p;
      } else {
	q = m->m_keywords;
	do {
	  run = 0;
	  c = strcmp(p->k_name, q->k_name);
	  if (c > 0) {
	    if (q->k_right == NULL) {
	      q->k_right = p;
	    } else {
	      q = q->k_right;
	      run++;
	    }
	  } else if (c < 0) {
	    if (q->k_left == NULL) {
	      q->k_left = p;
	    } else {
	      q = q->k_left;
	      run++;
	    }
	  } else {
	    mex_delkeyword(p);
	    result++;
	  }
	} while (run);
      }

    } else {
      free(p);
      result++;
    }
  } else {
    result++;
  }

  return result;
}

char *mex_error(MEX * m)
{
  static char result[MEX_BUFFER];

  result[0] = '\0';

  switch (m->m_error) {
  case MEX_ERROR_OK:
    strncpy(result, "no error", MEX_BUFFER - 1);
    break;
  case MEX_ERROR_TOOLONG:
    snprintf(result, MEX_BUFFER, "token near line %d is too long",
	     m->m_line);
    break;
  case MEX_ERROR_READ:
    strncpy(result, "error on reading input", MEX_BUFFER - 1);
    break;
  case MEX_ERROR_MEMORY:
    strncpy(result, "unable to allocate memory", MEX_BUFFER - 1);
    break;
  case MEX_ERROR_LEX:
    strncpy(result, "lexical error", MEX_BUFFER - 1);
    break;
  }

  return result;
}

MEX_TOKEN *mex_get(MEX * m)
{
  MEX_TOKEN *t, *u;
  int rr;

  if (m->m_token) {
    t = m->m_token;
    m->m_token = t->t_next;

    /* keep token around to collect */
    t->t_next = m->m_garbage;
    m->m_garbage = t;
  } else {

    if (m->m_lexed > 0) {	/* shuffle processed stuff out of the way */
#ifdef TRACE
      fprintf(stderr, "mex_get(): moving <%d> from <%d> to <0>\n",
	      m->m_read - m->m_lexed, m->m_lexed);
#endif
      if (m->m_read > m->m_lexed) {
	memmove(m->m_buf, m->m_buf + m->m_lexed, m->m_read - m->m_lexed);
      }
      m->m_read = m->m_read - m->m_lexed;
      m->m_lexed = 0;
    }
    if ((m->m_read < MEX_BUFFER) && (m->m_eof == 0)) {	/* top up buffer */
      rr = read(m->m_fd, m->m_buf + m->m_read, MEX_BUFFER - m->m_read);
#ifdef TRACE
      fprintf(stderr, "mex_get(): read <%d>\n", rr);
#endif
      switch (rr) {
      case -1:
	switch (errno) {
	case EAGAIN:
	case EINTR:
	  /* can recover from these errors */
	  break;
	default:
	  m->m_error = MEX_ERROR_READ;
	  m->m_eof = 1;
	  break;
	}
	break;
      case 0:
	m->m_eof = 1;
	break;
      default:
	m->m_read = m->m_read + rr;
	break;
      }
    }
    if (m->m_error) {		/* in case of error, don't return token */
      t = NULL;
    } else {
      u = m->m_token;
      do {			/* grab several tokens in one go */
#ifdef TRACE
	fprintf(stderr, "mex_get(): lexed %d, read %d, max %d\n",
		m->m_lexed, m->m_read, MEX_BUFFER);
#endif
	t = mex_token(m);
	if (t != NULL) {
	  if (u) {
	    u->t_next = t;
	  } else {
	    m->m_token = t;
	  }
	  u = t;
	}
      } while (t != NULL);

      t = m->m_token;		/* hand back first token, if there is one */
      if (m->m_token) {
	m->m_token = t->t_next;
	/* keep token around to collect */
	t->t_next = m->m_garbage;
	m->m_garbage = t;
      }
    }

  }

  return t;
}

void mex_unget(MEX * m, MEX_TOKEN * t)
{
  if (t == NULL) {		/* probably unwise, but what the heck */
    t = m->m_garbage;
  }
  if (t && (t == m->m_garbage)) {
    m->m_garbage = t->t_next;
    t->t_next = m->m_token;
    m->m_token = t;
  }
}

static void mex_free(MEX_TOKEN * t)
{
  if (t->t_buf != NULL) {
    free(t->t_buf);
    t->t_buf = NULL;
  }
  free(t);
}

static MEX_TOKEN *mex_new(int id, int type, char *buf, int len, int line)
{
  MEX_TOKEN *t;
  t = malloc(sizeof(MEX_TOKEN));
  if (t) {
    t->t_id = id;
    t->t_type = type;
    t->t_next = NULL;

    if (buf && (len > 0)) {
      t->t_buf = malloc(sizeof(char) * (len + 1));
      if (t->t_buf) {
	memcpy(t->t_buf, buf, len);
	t->t_len = len;
	t->t_line = line;
	t->t_buf[len] = '\0';
#ifdef TRACE
	fprintf(stderr, "mex_new(): created token <%s:%d>\n", t->t_buf,
		t->t_len);
#endif
      } else {
	free(t);
	t = NULL;
      }
    } else {
      t->t_buf = NULL;
      t->t_len = 0;
    }
  }
  return t;
}

static MEX_TOKEN *mex_token(MEX * m)
{
  MEX_TOKEN *t = NULL;
  char *ptr;
  int i, j, l;
  int state;

  ptr = m->m_buf;
  i = m->m_lexed;
  j = m->m_lexed;
  l = m->m_read;
  state = MEX_STATE_EATWS;

  do {
    if (i >= l) {
      state = MEX_STATE_QUIT;
    } else {
      switch (state) {
      case MEX_STATE_EATWS:
	if (isspace(ptr[i])) {
	  if (ptr[i] == '\n') {
	    m->m_line = m->m_line + 1;
	  }
	  i++;
	  m->m_lexed = i;
	  /* stay in current state */
	} else if (m->m_keychars[(int) ptr[i]]) {
	  t =
	      mex_new(m->m_keychars[(int) ptr[i]], MEX_KEYWORD, ptr + i, 1,
		      m->m_line);
	  if (t == NULL) {
	    m->m_error = MEX_ERROR_MEMORY;
	  }
	  i++;
	  m->m_lexed = i;
	  state = MEX_STATE_QUIT;
	} else if (ptr[i] == '"') {
	  i++;
	  j = i;
	  state = MEX_STATE_STRING;
	} else if (ptr[i] == '#') {
	  i++;
	  state = MEX_STATE_EATHASH;
	} else {
	  j = i;
	  state = MEX_STATE_WORD;
	}
	break;
      case MEX_STATE_EATHASH:
	switch (ptr[i]) {
	case '\n':
	  i++;
	  m->m_line = m->m_line + 1;
	  state = MEX_STATE_EATWS;
	  m->m_lexed = i;
	  break;
	default:
	  i++;
	  break;
	}
	break;
      case MEX_STATE_WORD:
	if (isspace(ptr[i]) || m->m_keychars[(int) ptr[i]] || ptr[i] == '"'
	    || ptr[i] == '#') {
	  t = mex_new(0, MEX_WORD, ptr + j, i - j, m->m_line);
	  if (t == NULL) {
	    m->m_error = MEX_ERROR_MEMORY;
	  } else {
	    t->t_id = mex_findword(m, t->t_buf);
	    if (t->t_id != 0) {
	      t->t_type = MEX_KEYWORD;
	    }
	  }
	  m->m_lexed = i;
	  state = MEX_STATE_QUIT;
	} else {
	  i++;
	}
	break;
      case MEX_STATE_STRING:
	switch (ptr[i]) {
	case '"':
	  t = mex_new(0, MEX_STRING, ptr + j, i - j, m->m_line);
	  if (t == NULL) {
	    m->m_error = MEX_ERROR_MEMORY;
	  }
	  i++;
	  m->m_lexed = i;
	  state = MEX_STATE_QUIT;
	  break;
	case '\\':
	  if (i > j) {
#ifdef TRACE
	    fprintf(stderr,
		    "mex_token(): moving %d one forward ptr[%d]='%c' \n",
		    i - j, j, ptr[j]);
#endif
	    memmove(ptr + j + 1, ptr + j, i - j);
	  }
	  j++;
	  i++;

	  switch (ptr[i]) {
	  case 'a':
	    ptr[i] = '\a';
	    break;
	  case 'f':
	    ptr[i] = '\f';
	    break;
	  case 'n':
	    ptr[i] = '\n';
	    break;
	  case 'r':
	    ptr[i] = '\r';
	    break;
	  case 't':
	    ptr[i] = '\t';
	    break;
	  case 'v':
	    ptr[i] = '\v';
	    break;
	  }

	  i++;
	  break;
	case '\n':
	  m->m_line = m->m_line + 1;
	  /* fall */
	default:
	  i++;
	  break;
	}
	break;
      }
    }

  } while (state != MEX_STATE_QUIT);

  return t;
}

static void mex_delkeyword(MEX_KEY * k)
{
  if (k) {
    mex_delkeyword(k->k_left);
    mex_delkeyword(k->k_right);

    k->k_left = NULL;
    k->k_right = NULL;

    free(k->k_name);
    k->k_name = NULL;

    free(k);
  }
}

static int mex_findword(MEX * m, unsigned char *s)
{
  int result = 0;
  MEX_KEY *k;
  int c, run;

  k = m->m_keywords;
  run = 1;
  do {
    if (k) {
      c = strcmp(s, k->k_name);
      if (c > 0) {
	k = k->k_right;
      } else if (c < 0) {
	k = k->k_left;
      } else {
	result = k->k_id;
	run = 0;
      }
    } else {
      run = 0;
    }
  } while (run);

  return result;
}

static void mex_dumpkey(MEX_KEY * k, FILE * f, int p)
{
  int i;
  for (i = 0; i < p; i++) {
    fputc(' ', f);
  }
  fprintf(f, "%s:%d\n", k->k_name, k->k_id);
  if (k->k_left)
    mex_dumpkey(k->k_left, f, p + 2);
  if (k->k_right)
    mex_dumpkey(k->k_right, f, p + 2);
}

void mex_dump(MEX * m, FILE * f)
{
  MEX_TOKEN *t;
  int i;

  fprintf(f, "fd %d, line %d, read %d, lexed %d\n", m->m_fd, m->m_line,
	  m->m_read, m->m_lexed);
  fprintf(f, "Buffered tokens:");
  t = m->m_token;
  while (t != NULL) {
    fprintf(f, "%d:[%s:%d] ", t->t_id, t->t_buf, t->t_len);
    t = t->t_next;
  }
  fprintf(f, "\n");

  fprintf(f, "Key Words\n");
  if (m->m_keywords) {
    mex_dumpkey(m->m_keywords, f, 0);
  }
  fprintf(f, "Key Characters\n");
  for (i = 0; i < MEX_CHAR; i++) {
    if (m->m_keychars[i]) {
      fprintf(f, "%c ", i);
    }
  }
  fprintf(f, "\n");
}
